import requests
import json
import logging

from KicadModTree import *
from .footprint_handlers import *


def create_footprint(
	footprint_component_uuid, component_id, footprint_lib, output_dir, model_path_relative
):
	logging.info("creating footprint ...")

	# fetch the compoennt data for easyeda library
	response = requests.get(
		f"https://easyeda.com/api/components/{footprint_component_uuid}"
	)
	if response.status_code == requests.codes.ok:
		data = json.loads(response.content.decode())
	else:
		logging.error(
			f"create_footprint error. Requests returned with error code {response.status_code}"
		)
		return ()
	footprint_shape = data["result"]["dataStr"]["shape"]

	footprint_name, datasheet_link, assembly_process = get_footprint_info(component_id)

	# init kicad footprint
	kicad_mod = Footprint(f'"{footprint_name}"')
	kicad_mod.setDescription(f"{footprint_name} footprint")  # TODO Set real description
	kicad_mod.setTags(f"{footprint_name} footprint")

	class footprint_info:
		def __init__(self, footprint_name, output_dir, footprint_lib):
			self.max_X, self.max_Y, self.min_X, self.min_Y = (
				-10000,
				-10000,
				10000,
				10000,
			)  # I will be using these to calculate the bounding box because the node.calculateBoundingBox() methode does not seems to work for me
			self.assembly_process = assembly_process
			self.footprint_name = footprint_name
			self.output_dir = output_dir
			self.footprint_lib = footprint_lib
			self.model_path_relative = model_path_relative

	footprint_info = footprint_info(
		footprint_name=footprint_name, output_dir=output_dir, footprint_lib=footprint_lib
	)

	# for each line in data : use the appropriate handler
	for line in footprint_shape:
		args = [i for i in line.split("~") if i]  # split and remove empty string in list
		model = args[0]
		logging.debug(args)
		if model not in handlers:
			logging.warning(f"footprint : model not in handler :  {model}")
			pass
		else:
			handlers.get(model)(args[1:], kicad_mod, footprint_info)

	# set general values
	kicad_mod.append(
		Text(
			type="reference",
			text="REF**",
			at=[(footprint_info.min_X + footprint_info.max_X) / 2, footprint_info.min_Y - 2],
			layer="F.SilkS",
		)
	)
	kicad_mod.append(
		Text(
			type="user",
			text="REF**",
			at=[(footprint_info.min_X + footprint_info.max_X) / 2, footprint_info.max_Y + 4],
			layer="F.Fab",
		)
	)
	kicad_mod.append(
		Text(
			type="value",
			text=footprint_name,
			at=[(footprint_info.min_X + footprint_info.max_X) / 2, footprint_info.max_Y + 2],
			layer="F.Fab",
		)
	)

	# translate the footprint to be centered around 0,0
	kicad_mod.insert(
		Translation(
			-(footprint_info.min_X + footprint_info.max_X) / 2,
			-(footprint_info.min_Y + footprint_info.max_Y) / 2,
		)
	)

	if not os.path.exists(f"{output_dir}/{footprint_lib}"):
		os.makedirs(f"{output_dir}/{footprint_lib}")

	# output kicad model
	file_handler = KicadFileHandler(kicad_mod)
	file_handler.writeFile(f"{output_dir}/{footprint_lib}/{footprint_name}.kicad_mod")
	logging.info(f"created '{output_dir}/{footprint_lib}/{footprint_name}.kicad_mod'")

	# return and datasheet link footprint name to be linked with the schematic
	return (f"{footprint_lib}:{footprint_name}", datasheet_link)


def get_footprint_info(component_id):

	# send request to get assembly process and datasheet_link
	request_data = """{}\"currentPage\":1,\"pageSize\":100,
				\"keyword\":\"{}\",
				\"searchSource\":\"search\",
				\"componentAttributes\":[]{}"
				""".format(
		"{", component_id, "}"
	)

	response = requests.post(
		url="https://jlcpcb.com/shoppingCart/smtGood/selectSmtComponentList",
		headers={"Content-Type": "application/json;charset=utf-8"},
		data=request_data,
	)

	if response.status_code == requests.codes.ok:
		response = json.loads(response.content.decode())
	else:
		logging.error(
			f"get_footprint_info request error. Requests returned with error code {response.status_code}"
		)
		return ()

	footprint_name = datasheet_link = assembly_process = None
	footprint_name = (
		response["data"]["componentPageInfo"]["list"][0]["componentModelEn"]
		.replace(" ", "_")
		.replace("/", "_")
	)

	component_list = response["data"]["componentPageInfo"]["list"]
	for component in component_list:
		if component["componentCode"] == component_id:
			component_lcscid = component["componentId"]

			response = requests.get(
				f"https://jlcpcb.com/shoppingCart/smtGood/getComponentDetail?componentLcscId={component_lcscid}"
			)
			if response.status_code == requests.codes.ok:
				component_data = json.loads(response.content.decode())["data"]
			else:
				logging.error(
					f"get_footprint_info error, could not retrieve component's data. Requests returned with error code {response.status_code}"
				)
				return ()

			datasheet_link = component_data["dataManualUrl"]
			assembly_process = component_data["assemblyProcess"]
			logging.debug(f"'get_footprint_info : component_data : {component_data}")
			break

	if not footprint_name:
		footprint_name = "NoName"
		logging.warning(
			"Could not retrieve components information from JLCPCB, default name 'NoName'."
		)
	if not datasheet_link:
		datasheet_link = ""
		logging.warning("Could not retrieve datasheet link from JLCPCB")
	if not assembly_process:
		assembly_process = ""

	return (footprint_name, datasheet_link, assembly_process)



import requests
import logging
import os
import re
from KicadModTree import *

wrl_header = """#VRML V2.0 utf8
#created by JLC2KiCad_lib using the JLCPCB library
#for more info see https://github.com/TousstNicolas/JLC2KICAD_lib
"""


def get_3Dmodel(component_uuid, footprint_info, kicad_mod, translationZ, rotation):
	logging.info("creating 3D model ...")

	response = requests.get(f"https://easyeda.com/analyzer/api/3dmodel/{component_uuid}")
	if response.status_code == requests.codes.ok:
		text = response.content.decode()
	else:
		logging.error("request error, no 3D model found")
		return ()
	translationX, translationY, translationZ = (
		0,
		0,
		float(translationZ) / 3.048,
	)  # foot to mm

	wrl_content = wrl_header

	# get material list
	pattern = "newmtl .*?endmtl"
	matchs = re.findall(pattern=pattern, string=text, flags=re.DOTALL)

	materials = {}
	for match in matchs:
		material = {}
		material_id = ""
		for value in match.split("\n"):
			if value[0:6] == "newmtl":
				material_id = value.split(" ")[1]
			elif value[0:2] == "Ka":
				material["ambientColor"] = value.split(" ")[1:]
			elif value[0:2] == "Kd":
				material["diffuseColor"] = value.split(" ")[1:]
			elif value[0:2] == "Ks":
				material["specularColor"] = value.split(" ")[1:]
			elif value[0] == "d":
				material["transparency"] = value.split(" ")[1]

		materials[material_id] = material

	# get vertices list
	pattern = "v (.*?)\n"
	matchs = re.findall(pattern=pattern, string=text, flags=re.DOTALL)

	vertices = []
	for vertice in matchs:
		vertices.append(
			" ".join([str(round(float(coord) / 2.54, 4)) for coord in vertice.split(" ")])
		)

	# get shape list
	shapes = text.split("usemtl")[1:]
	for shape in shapes:
		lines = shape.split("\n")
		material = materials[lines[0].replace(" ", "")]
		index_counter = 0
		link_dict = {}
		coordIndex = []
		points = []
		for line in lines[1:]:
			if len(line) > 0:
				face = [int(index) for index in line.replace("//", "").split(" ")[1:]]
				face_index = []
				for index in face:
					if index not in link_dict:
						link_dict[index] = index_counter
						face_index.append(str(index_counter))
						points.append(vertices[index - 1])
						index_counter += 1
					else:
						face_index.append(str(link_dict[index]))
				face_index.append("-1")
				coordIndex.append(",".join(face_index) + ",")
		points.insert(-1, points[-1])

		shape_str = f"""
Shape{{
	appearance Appearance {{
		material  Material 	{{ 
			diffuseColor {' '.join(material['diffuseColor'])} 
			specularColor {' '.join(material['specularColor'])}
			ambientIntensity 0.2
			transparency {material['transparency']}
			shininess 0.5
		}}
	}}
	geometry IndexedFaceSet {{
		ccw TRUE 
		solid FALSE
		coord DEF co Coordinate {{
			point [
				{(", ").join(points)}
			]
		}}
		coordIndex [
			{"".join(coordIndex)}
		]
	}}
}}"""

		wrl_content += shape_str

	if not os.path.exists(f"{footprint_info.output_dir}/{footprint_info.footprint_lib}"):
		os.makedirs(f"{footprint_info.output_dir}/{footprint_info.footprint_lib}")
	if not os.path.exists(
		f"{footprint_info.output_dir}/{footprint_info.footprint_lib}/packages3d"
	):
		os.makedirs(f"{footprint_info.output_dir}/{footprint_info.footprint_lib}/packages3d")

	filename = f"{footprint_info.output_dir}/{footprint_info.footprint_lib}/packages3d/{footprint_info.footprint_name}.wrl"
	with open(filename, "w") as f:
		f.write(wrl_content)

	if footprint_info.model_path_relative:
		dirname = ""
	else:
		dirname = os.getcwd().replace("\\", "/").replace("/footprint", "") + "/"

	kicad_mod.append(
		Model(
			filename=f"{dirname}{filename}",
			rotate=[-float(axis_rotation) for axis_rotation in rotation.split(",")],
		)
	)
	logging.info(f"added {filename} to footprint")




import logging
import re

RELATIVE_OFFSET = 0.1
ABSOLUTE_OFFSET_X = 40
ABSOLUTE_OFFSET_Y = -25


def h_R(data, kicad_schematic):
	"""
	Rectangle handler
	"""

	if len(data) == 12:
		X1 = float(data[0])
		Y1 = float(data[1])
		X2 = float(X1) + float(data[4])
		Y2 = float(Y1) + float(data[5])
	else:
		X1 = float(data[0])
		Y1 = float(data[1])
		X2 = float(X1) + float(data[2])
		Y2 = float(Y1) + float(data[3])

	X1 *= RELATIVE_OFFSET
	Y1 *= -RELATIVE_OFFSET
	X2 *= RELATIVE_OFFSET
	Y2 *= -RELATIVE_OFFSET

	X1 -= ABSOLUTE_OFFSET_X
	Y1 -= ABSOLUTE_OFFSET_Y
	X2 -= ABSOLUTE_OFFSET_X
	Y2 -= ABSOLUTE_OFFSET_Y

	kicad_schematic.drawing += f"""
      (rectangle
        (start {X1} {Y1})
        (end {X2} {Y2})
        (stroke (width 0) (type default) (color 0 0 0 0))
        (fill (type background))
      )"""


def h_E(data, kicad_schematic):
	"""
	Circle
	"""

	try:
		X1 = float(data[0]) * RELATIVE_OFFSET - ABSOLUTE_OFFSET_X
		Y1 = -float(data[1]) * RELATIVE_OFFSET - ABSOLUTE_OFFSET_Y
		radius = float(data[2]) * RELATIVE_OFFSET
		kicad_schematic.drawing += f"""
      (circle
        (center {X1} {Y1})
        (radius {radius})
        (stroke (width 0) (type default) (color 0 0 0 0))
        (fill (type background))
      )"""
	except:
		logging.exception("schematic : failed to add circle")


def h_P(data, kicad_schematic):
	"""
	Add Pin to the schematic
	"""

	if len(data) == 24:  # sometimes, the rotation parameter is not in the list.
		data.insert(5, "0")
	elif len(data) == 28:
		data.insert(1, "0")

	if data[1] == "0":
		electrical_type = "unspecified"
	elif data[1] == "1":
		electrical_type = "input"
	elif data[1] == "2":
		electrical_type = "output"
	elif data[1] == "3":
		electrical_type = "bidirectional"
	elif data[1] == "4":
		electrical_type = "power_in"
	else:
		electrical_type = "unspecified"

	pinNumber = data[2]
	pinName = data[13]

	X = float(data[3]) * RELATIVE_OFFSET - ABSOLUTE_OFFSET_X
	Y = -float(data[4]) * RELATIVE_OFFSET - ABSOLUTE_OFFSET_Y

	if data[5] in ["0", "90", "180", "270"]:
		rotation = (int(data[5]) + 180) % 360
	else:
		rotation = 0
		logging.warning(
			f'Schematic : pin number {pinNumber} : "{pinName}" failed to find orientation. Using Default orientation'
		)

	if rotation == 0 or rotation == 180:
		length = abs(float(data[8].split("h")[-1])) * RELATIVE_OFFSET
	elif rotation == 90 or rotation == 270:
		length = abs(float(data[8].split("v")[-1])) * RELATIVE_OFFSET

	try:
		if not kicad_schematic.pinNamesHide and data[9].split("^^")[1] == "0":
			kicad_schematic.pinNamesHide = "(pin_names hide)"
		if not kicad_schematic.pinNumbersHide and data[17].split("^^")[1] == "0":
			kicad_schematic.pinNumbersHide = "(pin_numbers hide)"

		nameSize = float(data[16].replace("pt", "")) * RELATIVE_OFFSET
		numberSize = float(data[24].replace("pt", "")) * RELATIVE_OFFSET
	except:
		nameSize = 0.6
		numberSize = 0.6

	kicad_schematic.drawing += f"""
      (pin {electrical_type} line
        (at {X} {Y} {rotation})
        (length {length})
        (name "{pinName}" (effects (font (size {nameSize} {nameSize}))))
        (number "{pinNumber}" (effects (font (size {numberSize} {numberSize}))))
      )"""


def h_T(data, kicad_schematic):
	"""
	Annotation handler
	"""

	try:
		mark = data[0]
		X = float(data[1]) * RELATIVE_OFFSET - ABSOLUTE_OFFSET_X
		Y = -float(data[2]) * RELATIVE_OFFSET - ABSOLUTE_OFFSET_Y
		angle = (float(data[3]) * 10 + 1800) % 3600

		fontSize = float(data[6].replace("pt", "")) * RELATIVE_OFFSET

		text = data[10]
		kicad_schematic.drawing += f"""
      (text
        "{text}"
        (at {X} {Y} {angle})
        (effects (font (size {fontSize} {fontSize})))
      )"""
	except:
		logging.exception("failed to add text to schematic")


def h_PL(data, kicad_schematic):
	"""
	Polygone handler
	"""

	try:
		pathString = data[0].split(" ")
		polypts = []
		for i, _ in enumerate(pathString[::2]):
			polypts.append(
				f"(xy {float(pathString[2*i]) * RELATIVE_OFFSET - ABSOLUTE_OFFSET_X} {-float(pathString[2*i+1]) * RELATIVE_OFFSET - ABSOLUTE_OFFSET_Y})"
			)
		polystr = "\n          ".join(polypts)

		kicad_schematic.drawing += f"""
      (polyline
        (pts
          {polystr}
        )
        (stroke (width 0) (type default) (color 0 0 0 0))
        (fill (type none))
      )"""
	except:
		logging.exception("Schematic : failed to add a polygone")


def h_PG(data, kicad_schematic):
	"""
	Closed polygone handler
	"""

	try:
		pathString = [i for i in data[0].split(" ") if i]
		polypts = []
		for i, _ in enumerate(pathString[::2]):
			polypts.append(
				f"(xy {float(pathString[2*i]) * RELATIVE_OFFSET - ABSOLUTE_OFFSET_X} {-float(pathString[2*i+1]) * RELATIVE_OFFSET - ABSOLUTE_OFFSET_Y})"
			)
		polypts.append(polypts[0])
		polystr = "\n          ".join(polypts)

		kicad_schematic.drawing += f"""
      (polyline
        (pts
          {polystr}
        )
        (stroke (width 0) (type default) (color 0 0 0 0))
        (fill (type background))
      )"""
	except:
		logging.exception("Schematic : failed to add a polygone")


def h_PT(data, kicad_schematic):
	"""
	Triangle handler
	"""

	data[0] = data[0].replace("M ", "").replace("L ", "").replace(" Z", "")
	h_PG(data, kicad_schematic)


def h_A(data, kicad_schematic):
	"""
	Arc handler
	"""

	from math import sqrt, acos, pi, sin, cos

	def getCenterParam(match):  # please forgive me, I have sinned
		e = float([i for i in re.split(r" |,", match[0][1]) if i][0])
		t = float([i for i in re.split(r" |,", match[0][1]) if i][1])
		s = float([i for i in re.split(r" |,", match[1][1]) if i][0])
		l = float([i for i in re.split(r" |,", match[1][1]) if i][1])
		r = float([i for i in re.split(r" |,", match[1][1]) if i][3])
		o = float([i for i in re.split(r" |,", match[1][1]) if i][4])
		n = float([i for i in re.split(r" |,", match[1][1]) if i][5])
		a = float([i for i in re.split(r" |,", match[1][1]) if i][6])

		def c(e, t, n, a):
			i = e * n + t * a
			r = sqrt((e * e + t * t) * (n * n + a * a))
			o = acos(i / r)
			return o

		u = 2 * pi
		g = 2 * pi
		h = 2 * pi
		m = 2 * pi
		f = 2 * pi
		if o < 0:
			o = -o
		if s < 0:
			s = -s
		if o == s:
			l = 0
		C = sin(l)
		y = cos(l)
		b = (e - n) / 2
		v = (t - a) / 2
		S = (e + n) / 2
		P = (t + a) / 2
		if o < 0.00001 or s < 0.00001:
			h = c(1, 0, n - e, a - t)
			return (S, P, h, pi)
		A = y * b + C * v
		T = y * v - C * b
		D = A * A / (o * o) + T * T / (s * s)
		if D > 1:
			o *= sqrt(D)
			s *= sqrt(D)
		k = o * s
		M = o * T
		I = s * A
		L = M * M + I * I
		if not L:
			return (S, P, 0, 0)
		w = (k * k - L) / L
		w = sqrt(abs(w))
		O = w * M / s
		R = -w * I / o
		u = y * O - C * R + S
		g = C * O + y * R + P
		E = (A - O) / o
		N = (A + O) / o
		F = (T - R) / s
		x = (T + R) / s
		h = c(1, 0, E, F)
		m = c(E, F, -N, -x)
		while m > f:
			m -= f
		while m < 0:
			m += f
		if r != 0:
			m -= f
		return (u, g, h, m)

	match = re.findall(r"([MA])([eE ,\-\+.\d]+)", data[0])
	cx, cy, theta, deltaTheta = getCenterParam(match)
	radius = float([i for i in re.split(r" |,", match[1][1]) if i][0])
	theta /= 2
	Xstart = cx + radius * cos(theta)
	Ystart = -(cy - radius * sin(theta))
	Xend = cx + radius * cos(theta + deltaTheta)
	Yend = -(cy - radius * sin(theta + deltaTheta))
	Xmid = cx + radius * cos(theta + deltaTheta / 2)
	Ymid = -(cy - radius * sin(theta + deltaTheta / 2))

	Xstart *= RELATIVE_OFFSET
	Ystart *= RELATIVE_OFFSET
	Xend *= RELATIVE_OFFSET
	Yend *= RELATIVE_OFFSET
	Xmid *= RELATIVE_OFFSET
	Ymid *= RELATIVE_OFFSET

	Xstart -= ABSOLUTE_OFFSET_X
	Ystart -= ABSOLUTE_OFFSET_Y
	Xend -= ABSOLUTE_OFFSET_X
	Yend -= ABSOLUTE_OFFSET_Y
	Xmid -= ABSOLUTE_OFFSET_X
	Ymid -= ABSOLUTE_OFFSET_Y

	kicad_schematic.drawing += f"""
      (arc
        (start {Xstart} {Ystart})
        (mid {Xmid} {Ymid})
        (end {Xend} {Yend})
        (stroke (width 0) (type default) (color 0 0 0 0))
        (fill (type none))
      )"""


handlers = {
	"R": h_R,
	"E": h_E,
	"P": h_P,
	"T": h_T,
	"PL": h_PL,
	"PG": h_PG,
	"PT": h_PT,
	"A": h_A,
	# "J" : h_NotYetImplemented,
	# "N" : h_NotYetImplemented,
	# "BE" : h_NotYetImplemented,
	# "AR" : h_NotYetImplemented,
	# "O" : h_NotYetImplemented,
}
